# Introduction

A web API server for serving meteo data that windPRO 4.0+ can import using the Meteo object GUI wizard.

This guide describes:

- How to set up and run the API server
- How to develop/adapt the API server to your needs

This guide is written with a focus on Linux (Debian/Ubuntu) since that is where the MeteoAPI is likely to be deployed,
but development on Windows is supported. (At EMD, the development happens on Windows, while the deployment has been
tested on Ubuntu 20.04 LTS.) If you want to test locally on Windows, you can still follow the guide (but ignore the
parts about Apache and mod_wsgi).

# Running the API server

## Set up Miniconda

### Windows

Download and install Miniconda from https://docs.conda.io/en/latest/miniconda.html.

Launch "Anaconda Powershell Prompt" from the Windows start menu, and keep using it for the rest of this guide.

Update conda, also if you just installed it two minutes ago:

- `conda update conda`

Install and configure conda-libmamba-solver (otherwise, conda will be very slow):

- `conda install conda-libmamba-solver`
- `conda config --set solver libmamba`

### Debian/Ubuntu

Download and install the Miniconda Python distribution and package manager (for example in your home directory):

- `wget https://repo.anaconda.com/miniconda/Miniconda3-latest-Linux-x86_64.sh`
- `sh Miniconda3-latest-Linux-x86_64.sh`
  - Anaconda recommends not installing as root, therefore no "sudo".
  - The installer will ask "Do you wish the installer to initialize Miniconda3 by running conda init?"  Answering
    "yes" will be a bit more convenient, but will make shell startup a bit slower. In the following I'll assume
    you answered "no".

Activate conda and its base environment for the current shell session (adapt the path to conda to where you
installed it):

- `eval "$(~/miniconda3/bin/conda shell.bash hook)"`
- The prompt string (the text to the left of the caret (cursor) on each new line in the shell) should now start
  with "(base) ".

Make sure you have the newest version of conda. Also do this if you just installed "Miniconda3-latest":

- `conda update conda`

Install and configure conda-libmamba-solver (otherwise, conda will be very slow):

- `conda install conda-libmamba-solver`
- `conda config --set solver libmamba`

Install conda-bash-completion to make conda more comfortable to work with in the shell:

- `conda install --channel conda-forge conda-bash-completion`
- Exit and restart the shell, and activate conda again (you can press Ctrl+R to search previously used commands,
  then write "hook" and press enter).

## Install Apache (optional)

For development and testing, the API server can be run with or without Apache+mod_wsgi. *Without* Apache+mod_wsgi,
you can run the API only locally on your developer machine. *With* Apache+mod_wsgi, you can run development and
testing on production-grade webserver software (Apache), either on your local machine, or on a dedicated
virtual/physical server. This guide will mention when the two approaches differ wrt. setting up the project.

If you want the ability to use Apache+mod_wsgi, you need to install the "apache2-dev" package system-wide now,
**before** installing mod_wsgi in the next section:

- `sudo apt install apache2-dev`

(Installing "apache2-dev" will also install "apache2" if it is not already installed.)

## Set up the virtual environment

Navigate to the directory containing the meteoapi code (same directory as this README.md):

- `cd /your/path/to/py_meteoapi`

Activate conda itself, if it is not already active (again, you can press Ctrl+R to search previously used commands,
then write "hook" and press enter).

Create the new conda environment:

- Either (if not using Apache+mod_wsgi):
  - `conda env create --name meteoapi --file environment.yml`
- Or (if using Apache+mod_wsgi):
  - `conda env create --name meteoapi --file environment.modwsgi.yml`

Activate the new conda environment:

- `conda activate meteoapi`

If using Apache+mod_wsgi, test that mod_wsgi-express works:

- `mod_wsgi-express start-server --log-to-terminal --startup-log --access-log --httpd-executable /usr/sbin/apache2
  --modules-directory /usr/lib/apache2/modules/ --port 8000`
- Use a web browser to access the site. If you are running the web browser from the same machine as mod_wsgi, you
  can just go to http://127.0.0.1:8000/. If running the web browser from another machine, you need the web server's
  hostname or address, such as http://10.10.68.32:8000/. The mod_wsgi test page should say "My web site runs on Malt
  Whiskey".
- Press Ctrl+C to stop the web server again.

## Set up the Django project

Run the following with the conda virtual environment active and py_meteoapi as the current working directory.

Copy the file `meteoapi/settings_example.py` to the location `meteoapi/settings.py`.

Prepare static files (images, css, etc.) for serving:

- `python manage.py collectstatic`

Create the database for the Django project:

- `python manage.py migrate`

Create an admin user account (which will be used to administer the user accounts that use the MeteoAPI from windPRO):

- `python manage.py createsuperuser`

## Start the Django project

If using django runserver:

- `python manage.py runserver`

Or, if using Apache+mod_wsgi:

- `mod_wsgi-express start-server --log-to-terminal --startup-log --access-log --httpd-executable /usr/sbin/apache2
  --modules-directory /usr/lib/apache2/modules/ --port 8000 --url-alias /static staticfiles meteoapi/wsgi.py`

(It is also possible to use "manage.py runmodwsgi", see
https://github.com/GrahamDumpleton/mod_wsgi/blob/develop/README.rst#using-mod_wsgi-express-with-django.)

## Log in to the running Django project

- Use a web browser to go to http://10.10.68.32:8000/admin (again, where 10.10.68.32 is replaced with your machine's IP
  address or hostname)
- There should be a DisallowedHost error message, saying, e.g., "You may need to add '10.10.68.32' to ALLOWED_HOSTS."
- Edit `meteoapi/settings.py` to add your host to the ALLOWED_HOSTS list, e.g., `ALLOWED_HOSTS = [ '10.10.68.32' ]`
- Restart the web server
- Use a web browser to go to http://10.10.68.32:8000/admin again
- Log in using the credentials you created for the superuser account
- You should now see the Django administration web interface, where you can add user accounts for the employees that
  are going to use the MeteoAPI from windPRO.
  - Avoid reusing passwords --- the help text under the password field shows a randomly generated password you can use.

## Set up for production

- Read and follow the Django Deployment checklist: https://docs.djangoproject.com/en/4.2/howto/deployment/checklist/
- To configure mod_wsgi for production, see the mod_wsgi readme, especially the section "Running mod_wsgi-express as
  root" https://github.com/GrahamDumpleton/mod_wsgi/blob/develop/README.rst
- If using Apache+mod_wsgi directly (i.e., without mod_wsgi-express setup-server), you must configure
  Apache for "WSGIPassAuthorization On" to pass auth headers from Apache to Django.
- Use HTTPS, otherwise the BasicAuth credentials will be sent unencrypted.

## HTTPS

With mod_wsgi-express, HTTPS can be used not just in production, but also during development and testing. The flags
`--ssl-certificate-file`, `--ssl-certificate-key-file`, `--https-port`, and `--server-name` need to be specified
when running mod_wsgi-express start-server or mod_wsgi-express setup-server. This requires a DNS domain name and a TLS
certificate for it.

If your server does not have a DNS domain name (yet), but you do have access to generate certificates from an internal
certicate authority (such as Active Directory Enterprise CA), you can use a domain name from nip.io which provides
DNS for IP addresses. For example, if your server has the internal IP address 10.10.68.32, you can access internally it
as the domain name 10.10.68.32.nip.io, and you can use your internal certicate authority to issue a TLS certificate for
this domain.

Here is an example of a full invocation of mod_wsgi-express with HTTPS:

- `mod_wsgi-express start-server --log-to-terminal --startup-log --access-log --httpd-executable /usr/sbin/apache2
  --modules-directory /usr/lib/apache2/modules/ --url-alias /static staticfiles meteoapi/wsgi.py --ssl-certificate-file
  ~/py_meteoapi_certs/10.10.68.32.nip.io.crt --ssl-certificate-key-file ~/py_meteoapi_certs/10.10.68.32.nip.io.key
  --https-only --https-port 8443 --server-name 10.10.68.32.nip.io`

# Developing the API server

## External documentation

The Meteo API server is built on Django and Django REST framework. You can find documentation and tutorials here:

- https://docs.djangoproject.com/en/4.2/
- https://www.django-rest-framework.org/

## The API design

The API consists of two main features: The ProjectListView and the DatasetDownloadView.

- ProjectListView provides windPRO with metadata about your measurement projects, measurement locations, and datasets
- DatasetDownloadView provides windPRO with the actual timeseries data for one dataset, packed into a single file.

### ProjectListView

This view returns metadata of this structure:

- Project
  - Location
    - Dataset

A project has one or more measurement locations, and a measurement location has one or more datasets, i.e., revisions of
the actual timeseries data. Datasets allow for having multiple versions of the same measurement location data, e.g., as
they are cleaned, and as new raw data is appended.

In the example implementation of the Meteo API server, this metadata is hardcoded. For a real usage scenario, this
data is likely to be read from a database. This could be an existing database you already have independently of the
Meteo API server, or you could use the Meteo API server's database for the purpose. See the section Database below.

The three types of metadata objects have various data fields. See dataclasses.py for details. Of note are the fields
"Created by" and "Modified by". These are strings, and it is up to you to decide what they should contain --- it
could be person names, initials, email addresses, or something else. It is possible that more fields will be added
in the future.

### DatasetDownloadView

This view returns a single data file to windPRO. Different file formats are supported by windPRO:

- windPRO meteo export
- windPRO mesores
- Plaintext timeseries data

The example implementation of the Meteo API server demonstrates all three, but if you are adapting the example code
to your needs, you probably only need to use one of the formats. And unless you already have wind data in the
windPRO-specific formats, you most likely only need to use Plaintext timeseries data.

Either way, the data file should use the format specified by the data_format property of the metadata Dataset object
that the timeseries data corresponds to. The single file that DatasetDownloadView returns to windPRO does not need to
have a filename, and if it does, the filename is not used by windPRO. Here are the details of all three formats:

For WINDPRO_METEO_EXPORT, the file should be a windPRO meteo export (plaintext) file. The position of the meteo
object in windPRO will be moved to the coordinates contained in the file (not the API metadata location).

For WINDPRO_MESORES (https://www.emd-international.com/files/flow/EMD_technote_MESORES_20230426.pdf), the file
should be a windPRO mesores file containing only one geographical location (if there is more than one, only the
first location will be read by windPRO when used with Meteo API). The position of the meteo object in windPRO will
be moved to the coordinates contained in the file (not the API metadata location). The other data in the file,
besides meteo data, will also be imported.

For PLAINTEXT_TIMESERIES_DATA, the file should be a zip file containing one or more plaintext files. This zip
file is imported in the same way as the windPRO Meteo "GO time series" function. This means the names of the
plaintext files will be shown in the windPRO Meteo Import setup window, but otherwise don't matter. The plaintext
files can also be organized in folders inside the zip file, and the names and structure of the folders also don't
matter. All plaintext files in the zip file should use the same format (column separator, date format, time zone,
etc.). The position of the meteo object in windPRO will be moved to the coordinates from the
API metadata location (not coordinates contained in the file(s), because plaintext data is not one standardized format).

For PLAINTEXT_TIMESERIES_DATA also, the format insides the text files needs to be mapped to column types in windPRO.
If the format is recognized by windPRO (e.g., SymphoniePRO or Ammonit) it will be parsed and imported automatically.
Otherwise, windPRO will attempt to automatically detect the format and column data types, unless the zip file
containing the text files also contains, in its root, a file named precisely "importfilter.wlsx". If this file is
present, windPRO will use the import filter specified in this file. The documentation of the XML-based WLSX format
follows in the future.

In the example implementation of the Meteo API server, the timeseries data is included as example files. For a real
usage scenario, the timeseries data could similarly be read from a file system, or it could come from a database, or
anything else you can program. In case the data comes from a database, your code needs to transform the data into a
file that is sent from DatasetDownloadView to windPRO. (If relevant, this can be done in-memory using "BytesIO".)

## Database

As mentioned in the ProjectListView section, the Meteo API metadata served by the API to windPRO is likely read from
a database. If you already have a database with some kind of metadata about masts, you can read it from the database
and map it to the format windPRO expects in the ProjectListView. You can connect to your existing database using the
Django Object-Relational Mapper (see supported databases at https://docs.djangoproject.com/en/4.2/ref/databases/), or
any other way to connect to a database from Python.

See also: https://docs.djangoproject.com/en/4.2/howto/legacy-databases/

Either way, the API server also needs to store some data of its own for user accounts, authentication, etc. By default,
the API server does this using SQLite, which stores all database data in a single file on the server filesystem. It is
possible to change this and have the API server store its own data in one of your own, existing databases.

Finally, if you don't already have an existing database with meteo metadata, it is also possible to expand the use of
the API server's own database to contain this metadata and/or even timeseries data. In this case, you should use the
Django ORM and specify the data models (tables) in the models.py file. See the Django documentation.

### The Django admin interface

Being built on Django, the Meteo API server has an administration web interface known as the Django admin interface.
In the example implementation, it is only used to administer the user accounts, but it is possible to expand it to
be a "content management" site for your Meteo API metadata (Projects, Locations, Datasets), if this data is already
accessible to the Django application, i.e. defined in the models.py file.

See also: https://docs.djangoproject.com/en/4.2/ref/contrib/admin/

## The Browsable API

Being built on Django REST framework, the Meteo API server contains a Browsable API that you can use during testing and
development. Use a web browser to go to the root of the project, e.g. http://127.0.0.1:8000/
or http://10.10.68.32:8000/, to see it.

## Automated tests

The example implementation contains a few automated tests in tests.py. You can use and expand them when adapting the
code to your needs. The tests can be run using an IDE such as PyCharm, or in the shell, with the conda virtual
environment active and py_meteoapi as the current working directory:

- `python manage.py test`
